/*
  Codice originale (leggermente modificato per rientrare in 1kB di
  flash) e descrizione del progetto:
  https://hackaday.io/project/18952-simon-game-with-attiny13
 
  Copyright (c) 2016 Divadlo fyziky UDiF (www.udif.cz)
 
  Permission is hereby granted, free of charge, to any person
  obtaining a copy of this software and associated documentation
  files (the "Software"), to deal in the Software without
  restriction, including without limitation the rights to use,
  copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the
  Software is furnished to do so, subject to the following
  conditions:
 
  The above copyright notice and this permission notice shall be
  included in all copies or substantial portions of the Software.
 
  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
  OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
  WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
  OTHER DEALINGS IN THE SOFTWARE.
 
 
  **** COMMENTI E SPIEGAZIONI INTRODUTTIVE ****
 
  SCOPO DEL PROGRAMMA
  Il codice implementa il classico gioco Simon con 4 LED, 4 pulsanti
  e una sequenza da ricordare; la sequenza casuale è generata quando 
  il pulsante start/reset avvia il gioco; la sequenza si allunga ogni
  volta che si indovina; il gioco si interrompe quando si sbaglia; il
  punteggio massimo viene salvto su EEPROM (si può cancellare premendo
  il pulsante del LED rosso durante lo start) 
 
  PREREQUISITI
  generalità sui microcontrollori e programmazione in C,
  saper leggere un datasheet lungo e complesso, bitwise operation e 
  bit masking per la gestione dei registri
 
  CONFIGURAZIONE
  Il MCU va configurato (con i "fuses") per andare a 1.2MHz (il ckock
  interno ha due possibili frequenze + un prescaler per ridurre la 
  frequenza) o l'esecuzione non sarà corretta come tempi
  Altre opzioni: BOD disabled, LTO enabled
 
  DATASHEET E REGISTRI PIU' IMPORTANTI UTILIZZATI
  I numeri di pagina nei commenti fanno riferimento al datasheet 
  dell'ATtiny13A. Nella sezione sui registri c'è il link alle pagine 
  con una descrizione e i seguenti valori di default:
  DDRB =   0b00000000 -> tutti ingressi
  PORTB =  0b00000000 -> pull-up disabilitato (tri-state)
  ADCSRA = 0b00000000 -> ADC spento
  ADMUX =  0b00000000 -> ADC0 selezionato
 
  DDRB imposta come uscite i bit a 1 (pagina 50)
  Tutti i pin sono ingressi tranne quando viene chiamata play(); il 
  pin 5 (PB5) viene usato come:
  * RESET attivo basso (start)
  * ADC0 scollegato  per generare il seed con l'ADC all'avvio
  Sul pin 6 (PB1/OC0B) si genera l'onda quadra per il buzzer usando il
  timer/counter e la modalità waveform generator.
  Per la gestione del timer/counter si usano i registri:
  - TCNT0 che contiene il valore del conteggio
  - OCR0A e OCR0B con dei valori da comparare con TCNT0 (valore 
  massimo che fissa la frequenza e valore in corrispondenza del
  quale c'è la commutazione tra livello alto e basso)
  Il waveform generator usa il risultato della comparazione per 
  produrre un segnale PWM nei pin OC0A e OC0B (collegato al buzzer)
  Il timer/clock si ferma se non si seleziona una sorgente di clock
  il modo di funzionamento si imposta con WGM0[2:0] e COM0x[1:0]:
  - normal: conta in su fino a 0xFF poi interrupt e ricomincia
  - CTC: conta fino a OCR0A
  - fast PWM: va basso su match con OCR0A alto su 0xFF (fig. 11-6)
  - phase correct PWM: conta in su e in giù
*/
 
 
// LIBRERIE
 
// sleep mode per ridurre il consumo. Al termine di ogni partita
// il MCU va in POWER-DOWN per non consumare e si sveglia col reset
#include <avr/sleep.h>
 
// come la funzione delay() di Arduino: cicli di CPU buttati ma va bene
//  per piccole temporizzazioni (altrimenti meglio usare il timer)
// il codice usa _delay_loop_2(t) dove t è un int a 16 bit
// il ritardo dipende dal clock del MCU e si calcola moltiplicando
// t per il tempo di quattro cicli di clock. Ad esempio con clk 1.2MHz 
// e t=25000 si calcola (1/clock)*4*t = 83ms
#include <util/delay_basic.h>
 
// per salvare l'high score nella EEPROM
#include <avr/eeprom.h>
 
// DICHIARAZIONE VARIABILI
 
// NB le variabili dichiarate e non inizializzate valgono zero, vedi
// www.microchip.com/webdoc/AVRLibcReferenceManual/FAQ_1faq_varinit.html
 
// i quattro elementi impostano come uscite PB1 (il buzzer) e uno tra
// PB3, PB2, PB0 e PB4 per accendere rispettivamente i LED 1, 4, 3 e 2
// nella fase di gioco in cui viene mostrata la sequenza
const uint8_t buttons[4] = {
  0b00001010, 0b00000110, 0b00000011, 0b00010010
};
 
// array con quattro valori usati dal waveform generator per produrre 
// le note: re# sol# do re# (2.5, 3.3, 4.2 e 5 kHz)
// vedere la funzione play() per interpretare i valori
const uint8_t tones[4] = {
  239, 179, 143, 119
};
 
uint8_t lastKey; // ultimo tasto premuto per il debouncing dei pulsanti
uint8_t lvl = 0; // livello (parte da 0 -> sequenza con un solo LED)
uint8_t maxLvl;  // livello massimo
 
// variabili per pseudo-random generator LFSR
uint16_t seed; // generato con l'ADC su un pin scollegato (e mischiato)
uint16_t ctx; // valori successivi generati a partire dal seed
 
// variabili volatile modificate dalla ISR del watchdog
volatile uint8_t nrot = 8; // quante volte si mischia il seed nella ISR
         // usando il valore del timer
volatile uint16_t time; // conteggio incrementato dall'ISR ogni 16ms
      // usato per il debouncing dei pulsanti e per
      // l'auto power-off dopo 64s (time = 4000)
 
// FUNZIONI 
 
void sleepNow() {
  // power-down riduce al massimo il consumo; il risveglio avviene col 
  // pulsant start/reset
  // di default tutti i pin sono ingressi; in power-down gli ingressi
  // flottanti è meglio che siano scollegati (pag. 53)
  PORTB = 0b00000000; // tri-state (disabilita la resitenza di pullup)
  cli(); // disabilita gli interrupt
  WDTCR = 0; // spegne il Watchdog timer
  set_sleep_mode(SLEEP_MODE_PWR_DOWN);
  sleep_enable();
  sleep_cpu();
}
 
void play(uint8_t i, uint16_t t = 45000) {
  // accende un LED e suona la nota corrisponente col buzzer
  // i è l'indice del LED da 0 a 3; t è usanto nella funzione delay
  // per ritardare di (4/1.2M)*t sec (45000 -> il delay è di 150ms)
 
  // con i pin impostati come ingressi modifico PORTB sapendo che:
  // 0 -> se è un input disabilita la resistenza di pull-up
  // 0 -> se è un output imposta il livello basso di tensione
  // questa operazione serve nel passaggio da ingresso con pull-up a
  // uscita ed è indicata a pagina 51
  PORTB = 0b00000000;
 
  // imposta come uscite il buzzer e uno dei pin collegati ai LED
  // con i bit di PORTB a 0 (livello basso) il LED si accende e sul
  // buzzer viene generato il segnale PWM 
  DDRB = buttons[i];  
 
  // assegno a OCR0A il valore massimo del conteggio del timer/counter
  // in modo da modificare la frequenza; da pag. 68 si capisce che la 
  // frequenza minima si otterrebbe senza usare OCR0A (quindi contando
  // sempre fino a 0xFF - 255) f = 1.2M/(255+255)= 1200000/510 = 2.3kHz
  // con i valori in tones si ottiene ad es. f=1.2M/(239+239)=2.5kHz
  // gli altri valori sono 3.3, 4.2 e 5 kHz (re# sol# do re#)
  OCR0A = tones[i];   // ad esempio 239 -> 11101111 per tones[0]
  // e a OCR0B la sua metà
  OCR0B = tones[i] >> 1; //shiftR di 1: 01110111 (239 diviso 2: 119)
 
  // TCCR0A/B sono impostati nel main prima di chiamare playa in modo
  // che il waveform generator funzioni in modalità PWM phase correct
  // WGM02 imposta il conteggio fino a OCR0A, CS01 il prescaler a 8
  TCCR0B = (1 << WGM02) | (1 << CS01);
 
  // suona la nota e tiene il LED acceso per 150ms
  _delay_loop_2(t);
 
  // spegne il timer/counter interrompendo il segnale PWM
  TCCR0B = 0b00000000; 
 
  // spegne il LED e disabilita il buzzer (tutti i pin come ingressi)
  DDRB = 0b00000000;
  // abilita le resistenze di pul-up per i pulsanti
  PORTB = 0b00011101;
}
 
void gameOver() {
  // animazione e gestione high-score
  for (uint8_t i = 0; i < 4; i++) {
    play(3 - i, 25000); // all'indietro
  }
  // se si batte l'high score salva il livello raggiunto su eeprom
  if (lvl > maxLvl) {
    eeprom_write_byte((uint8_t*) 0, ~lvl); // write best score
    //animazione high score (3 volte l'animazione level up)
    for (uint8_t i = 0; i < 3; i++) { 
      levelUp();
    }
  }
  sleepNow();
}
 
void levelUp() {
  // animazione con i quattro LED in sequenza
  for (uint8_t i = 0; i < 4; i++) {
    play(i, 25000);
  }
}
 
uint8_t simple_random4() {
  // per una spiegazione di come funziona vedi il progetto originale
  // o la pagina wikipedia (e l'AN della Maxim linkata):
  // it.wikipedia.org/wiki/Registro_a_scorrimento_a_retroazione_lineare
  // modifica ctx e restituisce un indice tra 0 e 3 per i LED
  // la sequenza pseudo-casuale di valori di ctx si ripete sempre 
  // uguale se non cambia il seed -> riportando ctx al valore del seed
  // si può ripetere la sequenza e confrontarla con la pressione dei 
  // pulsanti
  // Linear-feedback shift register
  // due volte perché servono due bit per un nuovo indice da 0 a 3
  for (uint8_t i = 0; i < 2; i++) {
    uint8_t lsb = ctx & 1; // estrae il LSB da ctx
    ctx >>= 1; // shift a destra
    if (lsb || !ctx) { // se LSB (uscita) è 1 o se ctx contiene tutti 0
                       // (con tutti zeri la sequenza non cambia mai)
      ctx ^= 0xB400; // inverte i bit secondo la maschera
    }
  }
  // l'AND con quella maschera dà il resto dopo aver diviso per 4
  // quindi un valore tra 0 e 3 che corrisponde al LED da accendere
  return ctx & 0b00000011;
}
 
ISR(WDT_vect) { // Watchdog Timeout Interrupt da avr-libc
  // eseguita ogni volta che scade il timer (16ms)
  // ogni 16ms viene incrementato il valore della variabile usata per 
  // gestire il poweroff dopo 64s di inattività e il debouncing
  time++; // increase each 16 ms
  // dopo ogni reset viene mischiato il seed per 8 volte, il seed non
  // cabia più fino al reset successivo
  if (nrot) {
    nrot--;
    // mischia il seed facendo uno shift a sinistra e XOR con valore
    // del timer/counter (8bit senza prescaler a 1.2MHz)
    seed = (seed << 1) ^ TCNT0;
  }
}
 
void resetCtx() {
  // riporta ctx al valore iniziale, quello del seed
  ctx = seed;
}
 
int main(void) {
  // pull-up sui 4 pulsanti dei LED (tutti ingressi di default)
  PORTB = 0b00011101;
 
  // GENERAZIONE DEL SEED
  // ADC
  // Il pin PB5/ADC0 (start/reset) viene usato per generare il seed 
  // attivando l'ADC quando è scollegato per ottenere un valore casuale
  // il valore viene poi mischiato otto volte nella ISR usando il timer
  ADCSRA |= (1 << ADEN); // abilita l'ADC (pagina 82 e 92)
  ADCSRA |= (1 << ADSC); // parte la conversione sul pin ADC0 scollegato
  // attende che termini la conversione ADC (ADSC vale 0 quando termina
  // la conversione (pagina 83)
  while (ADCSRA & (1 << ADSC)); 
  // l'ADC è un 4 canali a 10 bit ad approssimazioni successive; per 
  // ottenere 8 bit si considerano solo gli 8 bit meno significativi 
  // contenuti nel registro ADCL (gli altri due sono in ADCH)
  seed = ADCL;
  ADCSRA = 0b00000000; // spegne l'ADC
  // WATCHDOG TIMER
  // abilita l'interrupt del watchdog timer (interrupt mode pag 43)
  // il WDT usa un oscillatore separato a 128kHz (il timer va col clock)
  // senza prescaling c'è un timeout ogni 16ms (2048 cicli, pag 43) e
  // viene eseguita la ISR
  WDTCR = (1 << WDTIE); // parte il watchdog timer con prescaler da 16ms
  sei(); // global interrupt enable (mette a 1 il bit I nello SREG)
  // TIMER/COUNTER
  // imposta il timer/counter in normal mode (WGM[2:0] a zero) e senza 
  // prescaler (CS00 a uno): conta fino a 0xFF poi ricomincia e segnala
  // con TOV0 (pag. 64)
  // il timer comincia a contare quando si imposta una sorgente per il 
  // clock (pag. 60)
  TCCR0B = (1 << CS00);
 
  // il watchdog timer scade otto volte e la ISR mischia il seed 
  while (nrot); 
 
  // IMPOSTAZIONE TIMER/COUNTER per generare il segnale PWM del buzzer
 
  // cambia la modalità di funzionamento del timer che passa da normal 
  // a PWM phase correct (WGM0[2:0]=001) dove conta in su e in giù fino
  // a 0xFF (poi il valore massimo diventerà quello del reistro OCR0A 
  // nella funzione play() per cambiare la frequenza della nota)
  // Si imposta il compare output mode del waveform generator in modo 
  // che il pin del buzzer PB1/OC0B (COM0B[1:0]=10) vada basso sul
  // match col valore del registro OCR0B mentre conta in su e alto 
  // mentre conta in giù (OCR0B è impostato nella funzione play() come
  // OCR0A)
  // NB il segnale PWM è presente solo se il pin è impostato come usicta
  // (pag 68) quindi il buzzer non suona finche non si chiama play()
  TCCR0A = (1 << COM0B1) | (0 << COM0B0) | (0 << WGM01)  | (1 << WGM00); 
 
  // LETTURA O RESET DELL'HIGH SCORE DA EEPROM PREMENDO START+ROSSO
 
  // maxLvl viene scritto/letto negato (operatore ~) perché il valore
  // contenuto inizialmente nella EEPROM è sempre 0xFF (255 e non zero)
  // (uint8_t*) 0 è un puntatore a un byte che parte dall'indirizzo 0
  maxLvl = ~eeprom_read_byte((uint8_t*) 0); 
 
  // legge lo stato dei pin della port B e lo confronta con la maschera
  switch (PINB & 0b00011101) {
    // AND tra gli ingressi e la maschera sopra; 0 significa premuto, 
    // PB1 è il buzzer e PB5 non conta (è un reset attivo basso ma in 
    // questo punto del programma sarà già alto); i pin dei pulsanti 
    // non premuti sono al livello alto (pull-up) quindi controlliamo 
    // se PB3 (LED rosso) è premuto e nel caso resettiamo l'high score
    case 0b00010101: // PB3 premuto dopo il reset
      eeprom_write_byte((uint8_t*) 0, 255); // scrive 255 -> livello 0
      maxLvl = 0;
      break;
  }
 
  while (1) {
    // MOSTRA LA SEQUENZA
    // reimposta ctx al valore del seed; la sequenza di valori di ctx 
    // sarà sempre la stessa sequenza finché non si resetta
    resetCtx();
    // accende in sequenza un numero di LED pari al livello più uno
    for (uint8_t cnt = 0; cnt <= lvl; cnt++) { 
      // attesa tra un LED e il successivo; valore massimo 65536 (2^16)
      // -> 218ms poi scende se aumenta lvl
      _delay_loop_2(4400 + 489088 / (8 + lvl));
      // accende il LED con indice da 0 a 3 restituito da simple_random4()
      // generato come resto della divisione per 4 di ctx, il cui valore
      // viene aggiornato ad ogni chiamata
      play(simple_random4());
    }
 
    // INSERIMENTO SEQUENZA DA PARTE DEL GIOCATORE
 
    // azzera il conteggio che porta al poweroff
    time = 0;
    // imposta lastKey a un valore che non corrisponde a nessun pulsante
    lastKey = 5;
    // riporta ctx al valore iniziale uguale al seed
    resetCtx();
    // tante volte quanti sono i LED da indovinare
    for (uint8_t cnt = 0; cnt <= lvl; cnt++) {
      bool pressed = false;
      // finché non si preme un bottone
      while (!pressed) {
  // provo i quattro bottoni
        for (uint8_t i = 0; i < 4; i++) {
    // controlla se il pulsante i è premuto
    // buttons & maschera -> bit a 1 solo per il pin/pulsante i
    // PINB ha tutti i bit a 1 (resistenze di pull-up attivate
    // nella funzione play) tranne quello del pulsante premuto
    // l'espressione logica tra le parentesi interne vale 0 se
    // il pulsante i è premuto, poi viene complementata
          if (!(PINB & buttons[i] & 0b00011101)) {
            // DEBOUNCE
            // se sono passati 16ms (tempo max per eventuali rimbalzi o
      // se è stato premuto un pulsante diverso (non è rimbalzo)
            if (time > 1 || i != lastKey) {
              // accendi il LED i e registra la pressione del tasto
        play(i);
              pressed = true;
        // genera di nuovo l'indice LED della sequenza riprodotta
              uint8_t correct = simple_random4();
              // se si è premuto il pulsante sbagliato
        if (i != correct) {
    // accendi il LED giusto tante volte pari al livello raggiunto
                for (uint8_t j = 0; j < lvl; j++) {
                  _delay_loop_2(65536); // 33ms
                  play(correct, 45000); // 66ms
                }
                _delay_loop_2(65536); // 218ms
    // animazione game over
                gameOver();
              }
        // è stato premuto il pulsante giusto 
              time = 0; // azzera time
              lastKey = i; // memorizzo l'indice dell'ultimo LED
              break; // esce dal ciclo for e smette di controllare se è 
               // premuto un pulsante; passa al LED successivo 
         // della sequenza
            }
            time = 0; // azzera il conteggio per poweroff e debouncing
          }
        }
  // se passano 64s (time = 4000) e non succede niente poweroff
        if (time > 4000) {
          sleepNow();
        }
      }
    }
    _delay_loop_2(65536);
    // sequenza indovinata, aumenta il livello
    if (lvl < 254) {
      lvl++;
      levelUp(); // animazione levello completato
      _delay_loop_2(45000);
    }
  }
}
